@@ -1,16 +1,17 @@
 from __future__ import absolute_import, division, print_function
 
+import warnings
 import numpy as np
 
 from matplotlib.colors import Normalize
 from matplotlib.collections import LineCollection
 
-from mpl_scatter_density import ScatterDensityArtist
+from mpl_scatter_density.generic_density_artist import GenericDensityArtist
 
 from astropy.visualization import (ImageNormalize, LinearStretch, SqrtStretch,
                                    AsinhStretch, LogStretch)
 
-from glue.utils import defer_draw, broadcast_to, nanmax
+from glue.utils import defer_draw, broadcast_to, nanmax, ensure_numerical
 from glue.viewers.scatter.state import ScatterLayerState
 from glue.viewers.scatter.python_export import python_export_scatter_layer
 from glue.viewers.matplotlib.layer_artist import MatplotlibLayerArtist
@@ -64,9 +65,7 @@ def set_mpl_artist_cmap(artist, values, state=None, cmap=None, vmin=None, vmax=N
         vmin = state.cmap_vmin
         vmax = state.cmap_vmax
         cmap = state.cmap
-    if isinstance(artist, ScatterDensityArtist):
-        artist.set_c(values)
-    else:
+    if not isinstance(artist, GenericDensityArtist):
         artist.set_array(values)
     artist.set_cmap(cmap)
     if vmin > vmax:
@@ -152,9 +151,13 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
 
         # Scatter density
         self.density_auto_limits = DensityMapLimits()
-        self.density_artist = ScatterDensityArtist(self.axes, [], [], color='white',
-                                                   vmin=self.density_auto_limits.min,
-                                                   vmax=self.density_auto_limits.max)
+        with warnings.catch_warnings():
+            warnings.filterwarnings("ignore", message='All-NaN slice encountered')
+            self.density_artist = GenericDensityArtist(self.axes, color='white',
+                                                       vmin=self.density_auto_limits.min,
+                                                       vmax=self.density_auto_limits.max,
+                                                       update_while_panning=False,
+                                                       histogram2d_func=self.state.compute_density_map)
         self.axes.add_artist(self.density_artist)
 
         self.mpl_artists = [self.scatter_artist, self.plot_artist,
@@ -163,21 +166,22 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
         self.errorbar_index = 2
         self.vector_index = 3
 
-        self.reset_cache()
-
-    def reset_cache(self):
-        self._last_viewer_state = {}
-        self._last_layer_state = {}
+        # NOTE: Matplotlib can't deal with NaN values in errorbar correctly, so
+        # we need to prefilter values - the following variable is used to store
+        # the mask for the values we keep, so that we can apply it to the color
+        # See also https://github.com/matplotlib/matplotlib/issues/13799
+        self._errorbar_keep = None
 
     @defer_draw
-    def _update_data(self, changed):
+    def _update_data(self):
 
         # Layer artist has been cleared already
         if len(self.mpl_artists) == 0:
             return
 
         try:
-            x = self.layer[self._viewer_state.x_att].ravel()
+            if not self.state.density_map:
+                x = ensure_numerical(self.layer[self._viewer_state.x_att].ravel())
         except (IncompatibleAttribute, IndexError):
             # The following includes a call to self.clear()
             self.disable_invalid_attributes(self._viewer_state.x_att)
@@ -186,7 +190,8 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
             self.enable()
 
         try:
-            y = self.layer[self._viewer_state.y_att].ravel()
+            if not self.state.density_map:
+                y = ensure_numerical(self.layer[self._viewer_state.y_att].ravel())
         except (IncompatibleAttribute, IndexError):
             # The following includes a call to self.clear()
             self.disable_invalid_attributes(self._viewer_state.y_att)
@@ -196,12 +201,13 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
 
         if self.state.markers_visible:
             if self.state.density_map:
-                self.density_artist.set_xy(x, y)
+                # We don't use x, y here because we actually make use of the
+                # ability of the density artist to call a custom histogram
+                # method which is defined on this class and does the data
+                # access.
                 self.plot_artist.set_data([], [])
                 self.scatter_artist.set_offsets(np.zeros((0, 2)))
             else:
-                self.density_artist.set_xy([], [])
-                self.density_artist.set_c(None)
                 if self.state.cmap_mode == 'Fixed' and self.state.size_mode == 'Fixed':
                     # In this case we use Matplotlib's plot function because it has much
                     # better performance than scatter.
@@ -214,7 +220,6 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
         else:
             self.plot_artist.set_data([], [])
             self.scatter_artist.set_offsets(np.zeros((0, 2)))
-            self.density_artist.set_xy([], [])
 
         if self.state.line_visible:
             if self.state.cmap_mode == 'Fixed':
@@ -245,8 +250,8 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
 
             if self.state.vx_att is not None and self.state.vy_att is not None:
 
-                vx = self.layer[self.state.vx_att].ravel()
-                vy = self.layer[self.state.vy_att].ravel()
+                vx = ensure_numerical(self.layer[self.state.vx_att].ravel())
+                vy = ensure_numerical(self.layer[self.state.vy_att].ravel())
 
                 if self.state.vector_mode == 'Polar':
                     ang = vx
@@ -277,17 +282,28 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
 
         if self.state.xerr_visible or self.state.yerr_visible:
 
+            keep = ~np.isnan(x) & ~np.isnan(y)
+
             if self.state.xerr_visible and self.state.xerr_att is not None:
-                xerr = self.layer[self.state.xerr_att].ravel()
+                xerr = ensure_numerical(self.layer[self.state.xerr_att].ravel()).copy()
+                keep &= ~np.isnan(xerr)
             else:
                 xerr = None
 
             if self.state.yerr_visible and self.state.yerr_att is not None:
-                yerr = self.layer[self.state.yerr_att].ravel()
+                yerr = ensure_numerical(self.layer[self.state.yerr_att].ravel()).copy()
+                keep &= ~np.isnan(yerr)
             else:
                 yerr = None
 
-            self.errorbar_artist = self.axes.errorbar(x, y, fmt='none',
+            if xerr is not None:
+                xerr = xerr[keep]
+            if yerr is not None:
+                yerr = yerr[keep]
+
+            self._errorbar_keep = keep
+
+            self.errorbar_artist = self.axes.errorbar(x[keep], y[keep], fmt='none',
                                                       xerr=xerr, yerr=yerr)
             self.mpl_artists[self.errorbar_index] = self.errorbar_artist
 
@@ -304,11 +320,10 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
                 if self.state.cmap_mode == 'Fixed':
                     if force or 'color' in changed or 'cmap_mode' in changed:
                         self.density_artist.set_color(self.state.color)
-                        self.density_artist.set_c(None)
                         self.density_artist.set_clim(self.density_auto_limits.min,
                                                      self.density_auto_limits.max)
                 elif force or any(prop in changed for prop in CMAP_PROPERTIES):
-                    c = self.layer[self.state.cmap_att].ravel()
+                    c = ensure_numerical(self.layer[self.state.cmap_att].ravel())
                     set_mpl_artist_cmap(self.density_artist, c, self.state)
 
                 if force or 'stretch' in changed:
@@ -346,6 +361,7 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
 
                     if self.state.cmap_mode == 'Fixed':
                         if force or 'color' in changed or 'cmap_mode' in changed or 'fill' in changed:
+                            self.scatter_artist.set_array(None)
                             if self.state.fill:
                                 self.scatter_artist.set_facecolors(self.state.color)
                                 self.scatter_artist.set_edgecolors('none')
@@ -355,7 +371,7 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
                     elif force or any(prop in changed for prop in CMAP_PROPERTIES) or 'fill' in changed:
                         self.scatter_artist.set_edgecolors(None)
                         self.scatter_artist.set_facecolors(None)
-                        c = self.layer[self.state.cmap_att].ravel()
+                        c = ensure_numerical(self.layer[self.state.cmap_att].ravel())
                         set_mpl_artist_cmap(self.scatter_artist, c, self.state)
                         if self.state.fill:
                             self.scatter_artist.set_edgecolors('none')
@@ -370,7 +386,7 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
                             s = self.state.size * self.state.size_scaling
                             s = broadcast_to(s, self.scatter_artist.get_sizes().shape)
                         else:
-                            s = self.layer[self.state.size_att].ravel()
+                            s = ensure_numerical(self.layer[self.state.size_att].ravel())
                             s = ((s - self.state.size_vmin) /
                                  (self.state.size_vmax - self.state.size_vmin))
                             # The following ensures that the sizes are in the
@@ -393,7 +409,7 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
                 # Higher up we oversampled the points in the line so that
                 # half a segment on either side of each point has the right
                 # color, so we need to also oversample the color here.
-                c = self.layer[self.state.cmap_att].ravel()
+                c = ensure_numerical(self.layer[self.state.cmap_att].ravel())
                 self.line_collection.set_linearcolor(data=c, state=self.state)
 
             if force or 'linewidth' in changed:
@@ -409,7 +425,7 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
                     self.vector_artist.set_array(None)
                     self.vector_artist.set_color(self.state.color)
             elif force or any(prop in changed for prop in CMAP_PROPERTIES):
-                c = self.layer[self.state.cmap_att].ravel()
+                c = ensure_numerical(self.layer[self.state.cmap_att].ravel())
                 set_mpl_artist_cmap(self.vector_artist, c, self.state)
 
         if self.state.xerr_visible or self.state.yerr_visible:
@@ -420,7 +436,8 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
                     if force or 'color' in changed or 'cmap_mode' in changed:
                         eartist.set_color(self.state.color)
                 elif force or any(prop in changed for prop in CMAP_PROPERTIES):
-                    c = self.layer[self.state.cmap_att].ravel()
+                    c = ensure_numerical(self.layer[self.state.cmap_att].ravel()).copy()
+                    c = c[self._errorbar_keep]
                     set_mpl_artist_cmap(eartist, c, self.state)
 
                 if force or 'alpha' in changed:
@@ -458,30 +475,10 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
                 self.state.layer is None):
             return
 
-        # Figure out which attributes are different from before. Ideally we shouldn't
-        # need this but currently this method is called multiple times if an
-        # attribute is changed due to x_att changing then hist_x_min, hist_x_max, etc.
-        # If we can solve this so that _update_histogram is really only called once
-        # then we could consider simplifying this. Until then, we manually keep track
-        # of which properties have changed.
-
-        changed = set()
-
-        if not force:
-
-            for key, value in self._viewer_state.as_dict().items():
-                if value != self._last_viewer_state.get(key, None):
-                    changed.add(key)
-
-            for key, value in self.state.as_dict().items():
-                if value != self._last_layer_state.get(key, None):
-                    changed.add(key)
-
-        self._last_viewer_state.update(self._viewer_state.as_dict())
-        self._last_layer_state.update(self.state.as_dict())
+        changed = set() if force else self.pop_changed_properties()
 
         if force or len(changed & DATA_PROPERTIES) > 0:
-            self._update_data(changed)
+            self._update_data()
             force = True
 
         if force or len(changed & VISUAL_PROPERTIES) > 0:
@@ -497,3 +494,9 @@ class ScatterLayerArtist(MatplotlibLayerArtist):
     def update(self):
         self._update_scatter(force=True)
         self.redraw()
+
+    def remove(self):
+        super(ScatterLayerArtist, self).remove()
+        # Clean up the density artist to avoid circular references to do a
+        # reference to the self.histogram2d method in density artist.
+        self.density_artist = None
